"""
This class generates the query needed to resolved foreign keys for the ORM insert operation.

Author: Romke Jonker
Email: romke@stml.io
"""
from itertools import chain

from stimula.compiler.foreign_where_compiler import ForeignWhereClauseCompiler
from stimula.compiler.insert_compiler import FromClauseCompiler


class OrmInsertCompiler:
    """
        header: 'c1[unique=true], c2(c1), c3(b1), c4(b2(a1))'

        query:
        select c.c0, b.b0, b1.b0
        from c, b, b as b0
        join a on b0.b2=a.a0
        where c.c1 = :2 and b.b1 = :3 and a.a1=:4
    """

    def compile(self, mapping):
        select_clause = SelectClauseCompiler().compile(mapping)
        from_clause = FromClauseCompiler(False).compile(mapping)
        where_clause = ForeignWhereClauseCompiler(False, False).compile(mapping)

        # this where clause is also used in update query. Here we need 'where'
        if where_clause:
            where_clause = ' where ' + where_clause

        return f'{select_clause}{from_clause}{where_clause}'


class SelectClauseCompiler:
    def compile(self, mapping):
        # get list of attributes per column
        columns = [self._column(c) for c in mapping['columns']]
        # flatten list of lists
        attributes = [a for attributes in columns for a in attributes]
        # comma separate cells. For an insert, there may be nothing to select
        return 'select ' + ', '.join(attributes) if attributes else ''

    def _column(self, column):
        attributes = self._attributes(column['attributes'])
        return attributes

    def _attributes(self, attributes):
        # iterate attributes to get list of lists.
        return chain(*[self._attribute(a) for a in attributes])

    def _attribute(self, attribute):
        if 'foreign-key' in attribute:
            foreign_key = attribute['foreign-key']
            # but table names may need an alias
            target_name = foreign_key['table']
            target_alias = foreign_key.get('alias', target_name)
            target_column = foreign_key['name']
            # TODO: check if we need to alias the target column
            target_column_alias = target_column
            # get alias name from attribute.name, that's needed to get to res_id
            attribute_name = attribute['name']
            # no need to recurse
            return [f'{target_alias}.{target_column_alias} as {attribute_name}']

        # nothing to select for simple parameters
        return []


class OrmParameterNamesCompiler:
    # returns a list of parameter names that we need to pass to the orm create function, excluding
    # those that come from the query
    def compile(self, mapping):
        # get list of attributes per column
        return list(chain(*[self._column(c) for c in mapping['columns']]))

    def _column(self, column):
        return self._attributes(column['attributes'])

    def _attributes(self, attributes):
        # iterate attributes and flatten
        return chain(*[self._attribute(a) for a in attributes])

    def _attribute(self, attribute):
        if 'foreign-key' in attribute:
            # we'll resolve foreign keys in they query
            return []

        # return the parameter name, but without the colon
        return [f'{attribute["parameter"]}']



'''
Compiles foreign key where clauses for insert and update.
This compiler is used to filter the foreign key tables when inserting or updating records to obtain the foreign key values.
'''

from itertools import chain


class ForeignWhereClauseCompiler:

    def __init__(self, is_insert_query, is_delete_query):
        # if this is an insert query, we must not join with extension tables
        self._is_insert_query = is_insert_query
        self._is_delete_query = is_delete_query
        self._aliases = {}

    def compile(self, mapping):
        # glue cells together
        table_name = mapping['table']
        clauses = [self._column(c, table_name) for c in mapping['columns']]

        return ' and '.join(chain(*clauses))

    def _column(self, column, table):
        # no need to list non-unique columns as 'using' table when creating a delete query
        if self._is_delete_query and not column.get('unique', False):
            return []

        return self._attributes(column['attributes'], table, True)

    def _attributes(self, attributes, source_alias, is_root):
        return chain(*[self._attribute(attribute, source_alias, is_root) for attribute in attributes])

    def _attribute(self, attribute, alias, is_root_table):
        if not 'foreign-key' in attribute:

            # no where-clause needed for root table. Only add where clauses for joined tables, but do register for alias
            if is_root_table:
                return []

            # terminate foreign key with an equation
            parameter_name = attribute.get('parameter', f'{attribute["name"]}')
            return [f'{alias}.{attribute["name"]} = :{parameter_name}']

        foreign_key = attribute['foreign-key']
        target_name = foreign_key['table']
        target_alias = foreign_key.get('alias', target_name)

        # don't add extension conditions if this is an insert query
        if is_root_table and foreign_key.get('extension') and self._is_insert_query:
            return []

        # recurse
        where_clause = self._attributes(foreign_key['attributes'], target_alias, False)

        # add extension conditions if this is the root of an update or delete query
        if is_root_table and foreign_key.get('extension') and foreign_key.get('qualifier') and not self._is_insert_query:
            # assume for now that the alias is the table name. This is fine as long as we're not joining the same table multiple times
            model = alias.replace('_', '.')
            extension_clause = [f"{target_alias}.module = '{foreign_key['qualifier']}'",
                                f"{target_alias}.model = '{model}'"]
            where_clause = chain(where_clause, extension_clause)

        return where_clause



"""
This class generates delete statements based on a mapping.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from itertools import chain

from stimula.compiler.foreign_where_compiler import ForeignWhereClauseCompiler
from stimula.compiler.insert_compiler import ReturningClauseCompiler
from stimula.compiler.where_compiler import WhereClauseCompiler


class DeleteCompiler:
    """
        delete from c
        using c
        left join b on c.c0 = b.b0
        where c.c1 = :c1 and b.b1 = :b1;
    """

    def compile(self, mapping):
        table_name = mapping['table']
        using_clause = UsingClauseCompiler().compile(mapping)
        where_clause = WhereClauseCompiler().compile(mapping)
        foreign_where_clause = ForeignWhereClauseCompiler(False, True).compile(mapping)
        # in a delete statement, the foreign where clause always needs an 'and'
        foreign_where_clause = ' and ' + foreign_where_clause if foreign_where_clause else ''
        # returning clause is needed if we need to delete from an extension table
        returning_clause = ReturningClauseCompiler().compile(mapping)

        result = f'delete from {table_name}{using_clause}{where_clause}{foreign_where_clause}{returning_clause}'

        return result


class UsingClauseCompiler:
    # this class is similar to FromClauseCompiler, but it only uses unique columns
    def compile(self, mapping):
        # get table
        table_name = mapping['table']

        clause_lists = [self._column(c, table_name) for c in mapping['columns']]
        clauses = [clause for clause_list in clause_lists for clause in clause_list]

        if not clauses:
            return ''
        return f' using ' + ', '.join(clauses)

    def _column(self, column, table_name):
        # no need to list non-unique columns as 'using' table
        if not column.get('unique', False):
            return []

        return self._attributes(column['attributes'], table_name, True)

    def _attributes(self, attributes, source_alias, is_root):
        # get attributes that contains a foreign key
        foreign_key_attributes = [a for a in attributes if 'foreign-key' in a]

        return [self._attribute(attribute, source_alias, is_root) for attribute in foreign_key_attributes]

    def _attribute(self, attribute, source_alias, is_root_table):

        from_clause = ''

        # if not root, then assume we need a left join. This may require a modifier at some point.
        if not is_root_table:
            from_clause += ' left join '

        # get foreign key
        foreign_key = attribute['foreign-key']

        # add target table
        from_clause += foreign_key['table']

        # add alias
        target_name = foreign_key['table']
        target_alias = foreign_key.get('alias', target_name)
        if target_alias != target_name:
            from_clause += f' as {target_alias}'

        # if not root then add on clause
        if not is_root_table:
            from_clause += f' on {source_alias}.{attribute["name"]} = {target_alias}.{foreign_key["name"]}'

        # recurse
        return from_clause + ''.join(self._attributes(foreign_key['attributes'], target_alias, False))





'''
This class compiles a mapping into a where clause that does not recurse into foreign keys. Used for update and delete statements.

There's subtle difference between update/delete statements and the select statement that we need to find the primary key for ORM queries:
- update: update books set ... where ... book.authorid = authors.author_id
- delete: delete from books using authors where ... book.authorid = authors.author_id
- primary key selector: select ... from books join authors on book.authorid = authors.author_id

Author: Romke Jonker
Email: romke@stml.io
'''
from itertools import chain


class WhereClauseCompiler:

    def __init__(self, is_primary_key_selector=False):
        # if this is a primary key selector, we must not recurse into foreign keys
        self._is_primary_key_selector = is_primary_key_selector

    # This compiler create the usual where clause, but adds statements to restrict the 'using' tables
    def compile(self, mapping):
        table_name = mapping['table']
        # filter columns
        clause_lists = [self._column(c, table_name) for c in mapping['columns']]
        clauses = [clause for clause_list in clause_lists for clause in clause_list]

        # must not be empty, except for primary key selector, because that may be restricted by join instead
        if not clauses and not self._is_primary_key_selector:
            raise Exception('Header must have at least one unique column')

        return ' where ' + ' and '.join(clauses)

    def _column(self, column, table_name):
        unique = column.get('unique', False)
        if not unique:
            return []

        return chain(*[self._attribute(attribute, table_name) for attribute in column['attributes']])


    def _attribute(self, attribute, alias):
        # this compiler is only for the base table, don't recurse foreign keys
        column_name = attribute['name']
        if not 'foreign-key' in attribute:
            parameter_name = attribute.get('parameter', f'{column_name}')
            return [f'{alias}.{column_name} = :{parameter_name}']

        if self._is_primary_key_selector:
            # a primary key selector uses join...on instead of where...and, so don't create a where clause
            return []

        # update and delete statements must restrict by foreign keys
        foreign_key = attribute['foreign-key']
        target_name = foreign_key['table']
        target_alias = foreign_key.get('alias', target_name)
        target_column = foreign_key['name']
        return [f'{alias}.{column_name} = {target_alias}.{target_column}']
